import os
import subprocess
import sys
import pyDOE

import pandas as pd
import numpy as np
# import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt


sys.path.insert(0, 'D:\Git')
# sys.path.insert(0, '/home/tristan/git/')
from FDSInput.data_structures import FDSMESH
from FDSInput.data_structures import FDSOBST
from FDSInput.data_structures import FDSSURF
from FDSInput.data_structures import FDSMATL


# Location of experimental data files.
exp_data_dir = "ExpData"

# Data from the mass loss Cone experiments:
# -------------------------------------------
#
mlc_exp_file_labels = [["20_MLC_Alu",
                        "30_MLC_Alu",
                        "40_MLC_Alu",
                        "50_MLC_Alu",
                        "75_MLC_Alu"],
                       ["20_MLC_ISO",
                        "30_MLC_ISO",
                        "40_MLC_ISO",
                        "50_MLC_ISO",
                        "75_MLC_ISO"]]

# Labels for the different data series, corresponds to the keys in the
# `mlc_exp_info` dictionary.
mlc_col_labels = [["Time_Alu", "Mass_Norm_Alu", "SG_Alu"],
                  ["Time_ISO", "Mass_Norm_ISO", "SG_ISO"]]

# Dictionary collecting the experimental data file names, irradiance levels
# and rounded up durations of the longest test of each series.
mlc_exp_info = {'20_MLC_Alu': ['20_MLC_Alu.csv', 20.0, 1800, 300],
                '30_MLC_Alu': ['30_MLC_Alu.csv', 30.0, 1400, 300],
                '40_MLC_Alu': ['40_MLC_Alu.csv', 40.0, 1100, 300],
                '50_MLC_Alu': ['50_MLC_Alu.csv', 50.0, 1000, 300],
                '75_MLC_Alu': ['75_MLC_Alu.csv', 75.0, 700, 300],
                '20_MLC_ISO': ['20_MLC_ISO.csv', 20.0, 1600, 300],
                '30_MLC_ISO': ['30_MLC_ISO.csv', 30.0, 1200, 300],
                '40_MLC_ISO': ['40_MLC_ISO.csv', 40.0, 1000, 300],
                '50_MLC_ISO': ['50_MLC_ISO.csv', 50.0, 800, 300],
                '75_MLC_ISO': ['75_MLC_ISO.csv', 75.0, 600, 300]}


# List of parameter labels for easier access of the para_limits dictionary
# entries.
mlc_para_list = ['A',
                 'CONDUCTIVITY',
                 'DENSITY',
                 'E',
                 'EMISSIVITY',
                 'HEAT_OF_COMBUSTION',
                 'HEAT_OF_REACTION',
                 'N',
                 'SPECIFIC_HEAT']

mlc_para_resi_list = ['ResiEmis',
                      'ResiDens',
                      'ResiCond',
                      'ResiSpecHeat']

mlc_para_back_list = ['BackEmis',
                      'BackDens',
                      'BackCond',
                      'BackSpecHeat']


# Initialise a dictionary to create pandas DataFrames later on.
mlc_para_dict = {'EMISSIVITY': [],
                 'DENSITY': [],
                 'SPECIFIC_HEAT': [],
                 'A': [],
                 'E': [],
                 'HEAT_OF_COMBUSTION': [],
                 'HEAT_OF_REACTION': [],
                 'N': [],
                 'CONDUCTIVITY': [],
                 'file_number': []}

# Units of the used parameters.
unit_dict = {'EMISSIVITY': "--",
             'DENSITY': "kg$/$m$^3$",
             'SPECIFIC_HEAT': "kJ$/$(kg K)",
             'A': "1$/$s",
             'E': "kJ$/$kmol",
             'HEAT_OF_COMBUSTION': "kJ$/$kg",
             'HEAT_OF_REACTION': "kJ$/$kg",
             'N': "--",
             'CONDUCTIVITY': "W$/$(m K)"}

mlc_para_resi_dict = {'ResiEmis': [],
                      'ResiDens': [],
                      'ResiCond': [],
                      'ResiSpecHeat': []}

mlc_para_back_dict = {'BackEmis': [],
                      'BackDens': [],
                      'BackCond': [],
                      'BackSpecHeat': []}


# Original parameter values of the backing materials and residues.
# Note: Emissivity of 0.9 is FDS default.
mlc_para_resi_Alu_orig = {'ResiEmis': [0.9],
                          'ResiDens': [2698.9],
                          'ResiCond': [0.235],
                          'ResiSpecHeat': [0.897]}

mlc_para_back_Alu_dict = {'BackEmis': [0.9],
                          'BackDens': [2698.9],
                          'BackCond': [0.235],
                          'BackSpecHeat': [0.897]}

mlc_para_resi_ISO_orig = {'ResiEmis': [0.9],
                          'ResiDens': [2698.9],
                          'ResiCond': [0.235],
                          'ResiSpecHeat': [0.897]}

mlc_para_back_ISO_dict = {'BackEmis': [1.0],
                          'BackDens': [65],
                          'BackCond': [0.048],
                          'BackSpecHeat': [0.7]}


# Parameter sampling limits from the PROPTI IMP run.
mlc_para_limits = {'EMISSIVITY': [0.4, 0.99],
                   'DENSITY': [10, 2000],
                   'SPECIFIC_HEAT': [0.01, 2],
                   'A': [1e-1, 1e+3],
                   'E': [1e+2, 1e+6],
                   'HEAT_OF_COMBUSTION': [1e+4, 6e+4],
                   'HEAT_OF_REACTION': [100, 1e+4],
                   'N': [0.01, 1],
                   'CONDUCTIVITY': [0.01, 2]}


######################
# Information on the chaotic artificial bee (CAB) algorithm itself.
cab_info = {"Generation Size": {"CAB_1": [48, 96],
                                "CAB_2": [48, 96],
                                "CAB_3": [48, 96]}}



#####################################################
# Taken from SPOTPY
def mse(evaluation, simulation):
    """
    Mean Squared Error
        .. math::
         MSE=\\frac{1}{N}\\sum_{i=1}^{N}(e_{i}-s_{i})^2
    :evaluation: Observed data to compared with simulation data.
    :type: list
    :simulation: simulation data to compared with evaluation data
    :type: list
    :return: Mean Squared Error
    :rtype: float
    """

    if len(evaluation) == len(simulation):
        obs, sim = np.array(evaluation), np.array(simulation)
        mse = np.mean((obs - sim) ** 2)
        return mse
    else:
        logging.warning(
            "evaluation and simulation lists does not have the same length.")
        return np.nan


# Taken from SPOTPY
def rmse(evaluation, simulation):
    """
    Root Mean Squared Error
        .. math::
         RMSE=\\sqrt{\\frac{1}{N}\\sum_{i=1}^{N}(e_{i}-s_{i})^2}
    :evaluation: Observed data to compared with simulation data.
    :type: list
    :simulation: simulation data to compared with evaluation data
    :type: list
    :return: Root Mean Squared Error
    :rtype: float
    """
    if len(evaluation) == len(simulation) > 0:
        return np.sqrt(mse(evaluation, simulation))
    else:
        logging.warning(
            "evaluation and simulation lists do not have the same length.")
    return np.nan


# Taken from PROPTI
def map_to_def(x_def, x_values, y_values,
               factor=1.0, offset=0,
               target: str = 'model',
               mode: str = 'average',
               len_only: bool = False):
    """
    Maps the data of a data source to a definition set.

    :param target: choose data source, i.e. experiment or model,
        default: model, range: [model, experiment]
    :param mode: choose if data should be processed, default: average,
        range: [average]
    :param len_only: if set, return only the length of the resulting array,
        without creating the array
    :return: mapped data array or length of array
    """

    #     # set ds to none and if it stays none, something went wrong
    #     ds = None
    #     if target == 'model':
    #         ds = model
    #     if target == 'experiment':
    #         ds = experiment
    #     if ds is None:
    #         logging.error("* Wrong data read target: {}".format(target))
    #         sys.exit()

    # which mode?
    if mode == 'average':
        # if length is only required, return just the length of the
        # definition set
        if len_only:
            return len(x_def)

        # interpolate data on the definition set and return it
        return np.interp(x_def, x_values,
                         y_values) * factor + offset

    # wrong mode was chosen
    logging.error("* Wrong data mapping mode: {}".format(mode))
    sys.exit()


# Calculate RMSE
def calc_rmse(model_x, model_y,
              experiment_x, experiment_y):
    new_model_y = map_to_def(experiment_x, model_x, model_y)

    rmse_val = rmse(experiment_y, new_model_y)

    return rmse_val, new_model_y
#
#####################################################
