import os
import sys
import subprocess

import pandas as pd
import numpy as np

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker


# sys.path.insert(0, os.path.join("D", "Git"))
# # sys.path.insert(0, '/home/tristan/git/')
# from FDSInput.data_structures import FDSMESH
# from FDSInput.data_structures import FDSOBST
# from FDSInput.data_structures import FDSSURF
# from FDSInput.data_structures import FDSMATL


def exchange_placeholder(file_path, parameters):
    """
    Reads a text file from the provided path (file_path). Exchange
    parameters are taken from a list of lists (parameters). Each entry
    in the parameter list contains of a list with two elements, the
    first being the marker, the second the value. Markers will automatically
    encapsulated by '#' characters to provide compatibility to PROPTI.
    A list of lines will be returned in which the markers where exchanged to
    the values.

    :param file_path: (path to) a text file which leads to a simulation
        input file with markers
    :param parameters: list of lists of markers and values
        markers are looked for in the template and are exchanged
        to the values, format: [[marker, value]]
    :return: list of lists of the new input file content, line by line
    """
    new_file_content = []
    with open(file_path, 'r') as f:
        for line in f:
            for parameter in parameters:
                ph = '#{}#'.format(parameter[0])
                if ph in line:
                    line = line.replace(ph, str(parameter[1]))
                    # print(line)

            new_file_content.append(line)
    return new_file_content


def save_file(file_name, content, path, ending='fds',
              identifier=None, message=False):
    """

    :param file_name: name of the new file
    :param content: list of strings, list elements are lines
    :param path: where to write the file
    :param ending: file type, e.g. *.txt
    :param identifier: string, allows adding individual identifier to file name
    :param message: Suppress output

    :return: path as string
    """

    if identifier is not None:
        id_new = '_{}'.format(identifier)
    else:
        id_new = ''

    # Check if the directory already exists.
    # if not os.path.isdir(path):
    if not os.path.isdir(path):
        os.makedirs(path)
        if message is True:
            print("* New directory '{}' created.".format(path))
    else:
        if message is True:
            print("* Directory '{}' already exists.".format(path))

    # Create new file name and path.
    new_file_name = '{}{}.{}'.format(file_name, id_new, ending)
    new_file_path = os.path.join(path, new_file_name)

    # Check if the file already exists.
    if not os.path.isfile(new_file_path):
        with open(new_file_path, 'w') as f:
            for line in content:
                f.write(line)
        if message is True:
            print("* New file '{}' created.".format(new_file_name))
    else:
        if message is True:
            print("* File '{}' already exists.".format(new_file_name))

    # dir_list = next(os.walk(path))[1]
    # counter = 0
    # for directory in dir_list:
    #     counter += 1

    # new_dir = os.path.join(path, '{}{}'.format(file_name, id_new))
    # os.makedirs(new_dir)
    # print("* New sub-directory '{}' created.".format(new_dir))
    #
    # new_path = os.path.join(path, '{}{}.{}'.format(file_name, id_new, ending))

    # with open(new_file_path, 'w') as f:
    #     for line in content:
    #         f.write(line)

    return new_file_path


def run_job(file_path, message=False):
    """
    Provided with a path to a simulation input file, a sub-process
    is created and the simulation conducted.

    :param file_path: path to the simulation input file
        to be executed
    :param message: Suppress output
    """

    # Get the current working directory to switch between locations for
    # conducting simulations.
    cwd = os.getcwd()
    # print(os.path.split(file_path))

    # Split path and file name.
    # Path is needed to change directory.
    # File name is needed to launch the simulation, as well as looking for an
    #  *.end file to determine if the simulation has already been conducted,
    # assuming that the chid and file name are the same.
    wd, fn = os.path.split(file_path)
    end_file = fn[:-3] + "end"

    # Change directory for the execution of the simulation.
    os.chdir(wd)

    # Check if an *.end file exists to prevent running already completed
    # simulations.
    if not os.path.isfile(end_file):
        if message is True:
            print("* Run simulation...")
            print("  Job: '{}'".format(fn))

        subprocess.call("set OMP_NUM_THREADS=1 & fds {}".format(fn),
                        shell=True)

        if message is True:
            print("  Simulation completed.")
    else:
        if message is True:
            print("* End-file '{}' detected - simulation skipped.".format(
                end_file))

    # Return to the original working directory.
    os.chdir(cwd)


def box_plot(data_frame, generation_size, plot_name, out_dir,
             xlim0, xlim1, ylim0, ylim1, fig_x, fig_y):
    # Prepare list of lists to take the data for multiple boxplots, based
    # on the generations.
    data_series_multi = []

    # Calculate number of full generations. If last generation is
    # only partly complete it will be skipped.
    generations = len(data_frame["like1"])//generation_size
    print("Generations: ", generations)

    # Slice the column with the RMSE data according to generations.
    for i in range(generations):
        # Calculate where to start slicing (skip very first value).
        slice_begin = generation_size * i + 1
        # Calculate where to stop slicing (skip very first value).
        slice_end = generation_size * (i + 1) + 1
        # Slice data and convert into a list.
        new_data = data_frame['like1'][slice_begin:slice_end].tolist()
        # Collect the lists for later plotting.
        data_series_multi.append(new_data)

    # Prepare plotting of multiple boxplots in one diagramm.
    multi_plot = plt.figure(figsize=(fig_x, fig_y))
    # Call the subplots.
    ax = multi_plot.add_subplot(111)
    # Create multiple boxplots
    ax.boxplot(data_series_multi)

    # Finish the plot
    # pl.rcParams['figure.figsize'] = 16, 12
    # Prepare list for x-tick labels.
    x_tick_labels = ['', ]
    # Create x-tick labels based on number of generations.
    for i in range(generations):
        x_tick_labels.append(str(i))
    # Set values for the x-ticks.
    # plt.xticks(range(generations + 1),
    # plt.xticks(range(generations),
    #            x_tick_labels, rotation='horizontal')

    # # Set custom tick labels for every fifth tick.
    # ax.xaxis.set_major_locator(ticker.MultipleLocator(5))
    plt.xticks([1,5,10,15,20,25])
    tick_labels = ["1", "5", "10", "15", "20", "25"]
    ax.set_xticklabels(tick_labels)

    # plt.locator_params(axis='x', nbins=4)
    # plt.set(xticks=range(10), xlim=[xlim0, xlim1])
    # plt.xticks(range(10))



    plt. xlim(xlim0, xlim1)
    plt. ylim(ylim0, ylim1)

    plt.xlabel('Generations ({} individuals each)'.format(generation_size))
    plt.ylabel('Root Mean Square Error (RMSE)')
    # Create plot title from file name.
    # plt.title(plot_title)
    plt.grid()

    # Build file name and save.
    pfn = os.path.join(out_dir, plot_name)
    plt.savefig(pfn, dpi=320)


def plot_exp_mod_sim(title, x_label, y_label, data_set=None,
                     data_x_label=None, data_y_label=None,
                     exp_df=None, model_dfs=None,
                     exp_x_label=None, exp_y_label=None,
                     mod_x_label=None, mod_y_label=None,
                     x_min=0, x_max=600, y_min=0, y_max=700,
                     exp_y_fac=1, mod_y_fac=1, data_y_fac=1,
                     plot_file_name=None, fig_x=9, fig_y=8, dpi=320,
                     exp_color=['b'], mod_color=['k', 'r', 'g']):
    """

    :param data_set:
    :param title:
    :param x_label:
    :param y_label:
    :param data_x_label:
    :param data_y_label:
    :param exp_df:
    :param model_dfs:
    :param exp_x_label:
    :param exp_y_label:
    :param mod_x_label:
    :param mod_y_label:
    :param x_min:
    :param x_max:
    :param y_min:
    :param y_max:
    :param exp_y_fac:
    :param mod_y_fac:
    :param data_y_fac:
    :param plot_file_name:
    :param fig_x:
    :param fig_y:
    :param dpi:
    :param exp_color:
    :param mod_color:
    :return:
    """

    plt.figure(figsize=(fig_x, fig_y))

    # Plot experimental data.
    if exp_df is not None:
        plt.plot(exp_df[[exp_x_label]],
                 exp_df[[exp_y_label]] * exp_y_fac,
                 color=exp_color[0],
                 marker='.', linestyle='None')

    # Plot model data, related to above experiment.
    if model_dfs is not None:
        for i, model_df in enumerate(model_dfs):
            plt.plot(model_df[[mod_x_label]],
                     model_df[[mod_y_label]] * mod_y_fac,
                     color=mod_color[i])

    if data_set is not None:
        # Scale color cycle by number of data sets to plot.
        num_plots = len(data_set)
        plt.gca().set_prop_cycle(plt.cycler('color',
                                            plt.cm.viridis(np.linspace(0, 1,
                                                                       num_plots))))

        # Plot data sets, following the coloring definition above.
        for data_series in data_set:
            plt.plot(data_series[[data_x_label]],
                     data_series[[data_y_label]] * data_y_fac)

    plt.xlim(x_min, x_max)
    plt.ylim(y_min, y_max)

    plt.title(title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    plt.grid()

    if plot_file_name is not None:
        plt.savefig(plot_file_name + '.png', dpi=dpi)


def normaliser(x, xmin, xmax):
    """
    Normalises a parameters "location" within a given range. Lower range
    limit is set to 0, upper range limit is set to 1.

    :param x: Value to be normalised between the provided limits.
    :param xmin: Lower limit.
    :param xmax: Upper limit.
    :return: Normalised value.
    """

    normalised_value = (x-xmin)/(xmax-xmin)
    return normalised_value


def reverser(normal_value, xmin, xmax):
    """
    Maps normalised values back to the desired range.

    :param normal_value: Normalised value to be mapped onto the provided range.
    :param xmin: Lower limit.
    :param xmax: Upper limit.
    :return: Reversed value.
    """

    reversed_value = normal_value * (xmax - xmin) + xmin
    return reversed_value
